/*
 * Copyright (c) 2008, Harald Kuhr
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name "TwelveMonkeys" nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * Parts of this software is based on AC.lib-ICO.
 * See http://www.acproductions.de/commercial/aclibico/ for more information.
 *
 * The code is based on:
 *
 * winicontoppm.c - read a MS Windows .ico file and write portable pixmap(s)
 *
 * Copyright (C) 2000 by Lee Benfield - lee@recoil.org
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation. This software is provided "as is" without express or implied
 * warranty.
 *
 * Addendum for Java code by Christian Treber:
 *
 * The rules for the Java adaption are the same as stated above for the C
 * version.
 *
 * Copyright 2003 Christian Treber, ct@ctreber.com (Mar 27, 2003 17:00)
 */

package com.twelvemonkeys.imageio.plugins.ico;

import com.twelvemonkeys.image.InverseColorMapIndexColorModel;

import java.awt.image.BufferedImage;
import java.awt.image.DataBuffer;
import java.awt.image.IndexColorModel;
import java.awt.image.WritableRaster;
import java.util.Hashtable;

/**
 * Indexed (1, 4, and 8 bits per pixel) image.
 * <p/>
 *
 * @author <a href="ct@ctreber.com">Christian Treber</a>
 * @author <a href="mailto:harald.kuhr@gmail.com">Harald Kuhr</a>
 * @version $Id: ICOBitmapIndexed.java,v 1.0 25.feb.2006 00:29:44 haku Exp$
 */
class ICOBitmapIndexed extends ICOBitmap {
    protected final int[] mBits;
    protected final int[] mColors;

    public ICOBitmapIndexed(ICOEntry pEntry) {
        super(pEntry);
        mBits = new int[mWidth * mHeight];
        // NOTE: We're adding space for one extra color, for transparency
        mColors = new int[pEntry.getColorCount() + 1];
    }

    public BufferedImage createImageIndexed() {
        // TODO: This is very stupid, maybe we need a TYPE_CUSTOM image, with separate alphaRaster?!
        // As ICO has a separate bitmask, not related to palette index (allows 256 colors + trans) :-P

        IndexColorModel icm = createColorModel();

        Hashtable<String, Object> properties = null;
        if (mEntry instanceof CUREntry) {
            CUREntry entry = (CUREntry) mEntry;
            properties = new Hashtable<String, Object>(1);
            properties.put("cursor_hotspot", entry.getHotspot());
        }

        BufferedImage image = new BufferedImage(
                icm,
                icm.createCompatibleWritableRaster(mWidth, mHeight),
                icm.isAlphaPremultiplied(), properties
        );

        WritableRaster raster = image.getRaster();

        // Make pixels transparant according to mask
        final int trans = icm.getTransparentPixel();
        final ICOBitmapMask mask = mEntry.getMask();
        for (int y = 0; y < mHeight; y++) {
            for (int x = 0; x < mWidth; x++) {
                if (mask.getPaletteIndex(x, y) != 0) {
                    mBits[x + mWidth * y] = trans;
                }
            }
        }

        raster.setSamples(0, 0, mWidth, mHeight, 0, mBits);

        //System.out.println("Image: " + image);

        return image;
    }

    /**
     * @return Color model created from color palette in entry
     */
    IndexColorModel createColorModel() {
        // NOTE: This is a hack to make room for transparent pixel for mask
        int bits = mEntry.getBitCount();
        int colors = mColors.length;
        int trans = -1;

        // Try to avoid USHORT transfertype, as it results in BufferedImage TYPE_CUSTOM
        // NOTE: This code assumes icons are small, and is NOT optimized for performance...
        if (colors > (1 << mEntry.getBitCount())) {
            int index = ICOBitmapIndexed.findTransIndexMaybeRemap(mColors, mBits);

            if (index == -1) {
                // No duplicate found, increase bitcount
                bits++;
                trans = mColors.length - 1;
            }
            else {
                // Found a duplicate, use it as trans
                trans = index;
                colors--;
            }
        }

        // NOTE: Setting hasAlpha to true, makes things work on 1.2
        return new InverseColorMapIndexColorModel(
                bits, colors, mColors, 0, true, trans,
                bits <= 8 ? DataBuffer.TYPE_BYTE : DataBuffer.TYPE_USHORT
        );
    }

    private static int findTransIndexMaybeRemap(final int[] pColors, final int[] pBits) {
        // Look for unused colors, to use as transparent
        final boolean[] used = new boolean[pColors.length - 1];
        for (int pBit : pBits) {
            if (!used[pBit]) {
                used[pBit] = true;
            }
        }

        for (int i = 0; i < used.length; i++) {
            if (!used[i]) {
                return i;
            }
        }

        // Try to find duplicates in colormap, and remap
        int trans = -1;
        int duplicate = -1;
        for (int i = 0; trans == -1 && i < pColors.length - 1; i++) {
            for (int j = i + 1; j < pColors.length - 1; j++) {
                if (pColors[i] == pColors[j]) {
                    //System.out.println("Duplicate found: " + j);
                    trans = j;
                    duplicate = i;
                    break;
                }
            }
        }

        if (trans != -1) {
            // Remap duplicate
            for (int i = 0; i < pBits.length; i++) {
                if (pBits[i] == trans) {
                    pBits[i] = duplicate;
                }
            }
        }

        return trans;
    }

    public BufferedImage getImage() {
        if (mImage == null) {
            mImage = createImageIndexed();
        }
        return mImage;
    }

    protected int getColor(int pX, int pY) {
        return getColor(getPaletteIndex(pX, pY));
    }

    /**
     * @param pX x
     * @param pY y
     * @return Palette index for pixel x, y
     */
    protected int getPaletteIndex(int pX, int pY) {
        return mBits[pY * mWidth + pX];
    }

    protected int getColor(int pIndex) {
        return mColors[pIndex];
    }
}
