/*
 * Copyright (c) 2008, Harald Kuhr
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name "TwelveMonkeys" nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * Parts of this software is based on AC.lib-ICO.
 * See http://www.acproductions.de/commercial/aclibico/ for more information.
 *
 * The code is based on:
 *
 * winicontoppm.c - read a MS Windows .ico file and write portable pixmap(s)
 *
 * Copyright (C) 2000 by Lee Benfield - lee@recoil.org
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation. This software is provided "as is" without express or implied
 * warranty.
 *
 * Addendum for Java code by Christian Treber:
 *
 * The rules for the Java adaption are the same as stated above for the C
 * version.
 *
 * Copyright 2003 Christian Treber, ct@ctreber.com (Mar 27, 2003 17:00)
 */

package com.twelvemonkeys.imageio.plugins.ico;

import java.io.DataInput;
import java.io.IOException;
import java.awt.image.BufferedImage;

/**
 * Icon entry in image file.
 * <p/>
 *
 * @author <a href="ct@ctreber.com">Christian Treber</a>
 * @author <a href="mailto:harald.kuhr@gmail.com">Harald Kuhr</a>
 * @version $Id: ICOEntry.java,v 1.0 25.feb.2006 00:29:44 haku Exp$
 */
class ICOEntry {
    private ICOHeader mHeader;
    private int mWidth;
    private int mHeight;
    private int mColorCount;
    private int mReserved;
    int mPlanes;
    int mBitCount;
    private int mSize;
    private int mOffset;
    private ICOBitmap mBitmap;
    // RGB images have no mask
    private ICOBitmapMask mMask;

    // 16 Bytes
    public ICOEntry(DataInput pDec) throws IOException {
        read(pDec);
    }

    private void read(DataInput pDec) throws IOException {
        mWidth = pDec.readUnsignedByte();
        mHeight = pDec.readUnsignedByte();
        mColorCount = pDec.readUnsignedByte();
        mReserved = pDec.readUnsignedByte();
        mPlanes = pDec.readUnsignedShort();     // Always 1 for ICO, x hotspot for CUR
        mBitCount = pDec.readUnsignedShort();   // bit count for ICO, y hotspot for CUR
        mSize = pDec.readInt();
        mOffset = pDec.readInt();
    }

    public String toString() {
        return "Entry: width: " + mWidth +
                ", height: " + mHeight +
                ", colorCount: " + mColorCount + " (" + getColorCount() + ")" +
                ", planes: " + mPlanes +
                ", bitCount: " + mBitCount +
                ", size: " + mSize +
                ", offset: " + mOffset;
    }

    /**
     * @return Bits per pixel. If the bit count of the entry is 0, the bit count
     *         of the header is returned. See {@link #getBitCountRaw}.
     */
    public int getBitCount() {
        return mBitCount != 0 ? mBitCount : mHeader.getBitCount();
    }

    /**
     * @return The original bit count. See {@link #getBitCount}.
     */
    int getBitCountRaw() {
        return mBitCount;
    }

    /**
     * @return a BufferedImage. This method works for indexed color and
     *         RGB ICO files. Transparency information that might be present in
     *         the ICO is used.
     */
    public BufferedImage getImage() {
        return mBitmap.getImage();
    }

    /**
     * @return The original color count (note "0" means "256").
     *         See {@link #getColorCount}
     */
    int getColorCountRaw() {
        return mColorCount;
    }

    /**
     * @return The actual color count. See {@link #getColorCountRaw}
     */
    public int getColorCount() {
        return mColorCount == 0 && mBitCount <= 8 ? 256 : mColorCount;
    }

    public int getHeight() {
        return mHeight;
    }

    public int getOffset() {
        return mOffset;
    }

    public int getPlanes() {
        return mPlanes;
    }

    int getReserved() {
        return mReserved;
    }

    public int getSize() {
        return mSize;
    }

    public int getWidth() {
        return mWidth;
    }

    public ICOHeader getHeader() {
        return mHeader;
    }

    void setHeader(ICOHeader pHeader) {
        mHeader = pHeader;
    }

    public ICOBitmapMask getMask() {
        return mMask;
    }

    void setMask(ICOBitmapMask pMask) {
        mMask = pMask;
    }

    public ICOBitmap getBitmap() {
        return mBitmap;
    }

    void setBitmap(ICOBitmap pBitmap) {
        mBitmap = pBitmap;
    }
}
