/*
 * Copyright (c) 2008, Harald Kuhr
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name "TwelveMonkeys" nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * Parts of this software is based on AC.lib-ICO.
 * See http://www.acproductions.de/commercial/aclibico/ for more information.
 *
 * The code is based on:
 *
 * winicontoppm.c - read a MS Windows .ico file and write portable pixmap(s)
 *
 * Copyright (C) 2000 by Lee Benfield - lee@recoil.org
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation. This software is provided "as is" without express or implied
 * warranty.
 *
 * Addendum for Java code by Christian Treber:
 *
 * The rules for the Java adaption are the same as stated above for the C
 * version.
 *
 * Copyright 2003 Christian Treber, ct@ctreber.com (Mar 27, 2003 17:00)
 */

package com.twelvemonkeys.imageio.plugins.ico;

import javax.imageio.IIOException;
import java.io.DataInput;
import java.io.IOException;

/**
 * Icon header.
 * <p/>
 *
 * @author <a href="ct@ctreber.com">Christian Treber</a>
 * @author <a href="mailto:harald.kuhr@gmail.com">Harald Kuhr</a>
 * @version $Id: ICOHeader.java,v 1.0 25.feb.2006 00:29:44 haku Exp$
 */
class ICOHeader { // NOTE: This is actually a BMP header, the structure is exactly the same...
    // Always 40
    private int mHeaderSize;
    private int mWidth;
    // Weird, but this includes the height of the mask (so often header.height = entry.height * 2)
    private int mHeight;
    private int mPlanes;
    private int mBitCount;
    /**
     * 0 = BI_RGB: No compression
     * 1 = BI_RLE8: 8 bit RLE Compression (8 bit only)
     * 2 = BI_RLE4: 4 bit RLE Compression (4 bit only)
     * 3 = BI_BITFIELDS: No compression (16 & 32 bit only)
     */
    private int mCompression;
    // Can be 0 when compression == 0 (b/c size can be calculated then ?!)
    private int mImageSize;
    private int mXPixelsPerM;
    private int mYPixelsPerM;
    private int mColorsUsed;
    // 0: all
    private int mColorsImportant;

    // 40 Bytes + 4 * colors
    public ICOHeader(DataInput pDec) throws IOException {
        mHeaderSize = pDec.readInt();
        if (mHeaderSize != ICO.ENTRY_HEADER_SIZE) {
            throw new IIOException("Unknown header size: " + mHeaderSize);
        }
        mWidth = pDec.readInt();
        mHeight = pDec.readInt();
        mPlanes = pDec.readUnsignedShort();
        mBitCount = pDec.readUnsignedShort();
        mCompression = pDec.readInt();
        mImageSize = pDec.readInt();
        mXPixelsPerM = pDec.readInt();
        mYPixelsPerM = pDec.readInt();
        mColorsUsed = pDec.readInt();
        mColorsImportant = pDec.readInt();
    }

    public String toString() {
        return
                "Header: size: " + mHeaderSize +
                        ", width: " + mWidth +
                        ", height: " + mHeight +
                        ", planes: " + mPlanes +
                        ", bitCount: " + mBitCount +
                        ", compression: " + mCompression +
                        ", imageSize: " + mImageSize +
                        ", XPixelsPerM: " + mXPixelsPerM +
                        ", YPixelsPerM: " + mYPixelsPerM +
                        ", colorsUsed: " + mColorsUsed +
                        ", colorsImportant: " + mColorsImportant + (mColorsImportant == 0 ? " (all)" : "");
    }

    public int getBitCount() {
        return mBitCount;
    }

    public int getColorsImportant() {
        return mColorsImportant;
    }

    public int getColorsUsed() {
        return mColorsUsed;
    }

    public int getCompression() {
        return mCompression;
    }

    public int getHeight() {
        return mHeight;
    }

    public int getImageSize() {
        return mImageSize;
    }

    public int getPlanes() {
        return mPlanes;
    }

    public int getHeaderSize() {
        return mHeaderSize;
    }

    public int getWidth() {
        return mWidth;
    }

    public int getXPixelsPerM() {
        return mXPixelsPerM;
    }

    public int getYPixelsPerM() {
        return mYPixelsPerM;
    }
}
