/*
 * Copyright (c) 2008, Harald Kuhr
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name "TwelveMonkeys" nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * Parts of this software is based on AC.lib-ICO.
 * See http://www.acproductions.de/commercial/aclibico/ for more information.
 *
 * The code is based on:
 *
 * winicontoppm.c - read a MS Windows .ico file and write portable pixmap(s)
 *
 * Copyright (C) 2000 by Lee Benfield - lee@recoil.org
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation. This software is provided "as is" without express or implied
 * warranty.
 *
 * Addendum for Java code by Christian Treber:
 *
 * The rules for the Java adaption are the same as stated above for the C
 * version.
 *
 * Copyright 2003 Christian Treber, ct@ctreber.com (Mar 27, 2003 17:00)
 */

package com.twelvemonkeys.imageio.plugins.ico;

import com.twelvemonkeys.image.ImageUtil;
import com.twelvemonkeys.imageio.ImageReaderBase;
import com.twelvemonkeys.imageio.util.IndexedImageTypeSpecifier;

import javax.imageio.*;
import javax.imageio.metadata.IIOMetadata;
import javax.imageio.spi.ImageReaderSpi;
import javax.swing.*;
import java.awt.*;
import java.awt.color.ColorSpace;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.*;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * ImageReader for Windows ICO (icon) format.
 * Supports reading of all icons in one file.
 * 1, 4, 8 bit palette support with transparency, and 16, 24 and 32 bit
 * true color support with alpha.
 * <p/>
 * Parts of this ImageReader is based on a Java port of winicontoppm.c:
 * Copyright (C) 2000 by Lee Benfield - lee@recoil.org.
 * Java port:
 * Copyright 2003 Christian Treber, ct@ctreber.com.
 * 
 * @author <a href="ct@ctreber.com">Christian Treber</a>
 * @author <a href="mailto:harald.kuhr@gmail.com">Harald Kuhr</a>
 * @version $Id: ICOImageReader.java,v 1.0 25.feb.2006 00:29:44 haku Exp$
 *
 * @todo Support loading icons from DLLs, see
 * <a href="http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dnwui/html/msdn_icons.asp">MSDN</a>
 */
public class ICOImageReader extends ImageReaderBase {
    // TODO: Consider moving the reading to inner classes (subclasses of ICOBitmap)

    private ICOFile mICOFile;

    public ICOImageReader(int pType) {
        super(getProvider(pType));
    }

    ICOImageReader(ImageReaderSpi pProvider) {
        super(pProvider);
    }

    protected void resetMembers() {
        mICOFile = null;
    }

    private static ImageReaderSpi getProvider(final int pType) {
        switch (pType) {
            case ICO.TYPE_ICO:
                return new ICOImageReaderSpi();
            case ICO.TYPE_CUR:
                return new CURImageReaderSpi();
            default:
                throw new IllegalArgumentException("Unsupported type: " + pType);
        }
    }

    public int getHeight(int pImageIndex) throws IOException {
        return getICOEntry(pImageIndex).getHeight();
    }

    @Override
    public IIOMetadata getImageMetadata(int pImageIndex)
            throws IOException {
        return new ICOMetaData(getICOEntry(pImageIndex));
    }

    public Iterator<ImageTypeSpecifier> getImageTypes(int pImageIndex) throws IOException {
        List<ImageTypeSpecifier> types = new ArrayList<ImageTypeSpecifier>();
        ICOEntry entry = getICOEntry(pImageIndex);
        ICOHeader header = initEntryHeader(entry);

        // Use data from header to create specifier
        ImageTypeSpecifier specifier;
        switch (header.getBitCount()) {
            case 1:
            case 2:
            case 4:
            case 8:
                // TODO: This is QnD...
                int offset = entry.getOffset() + header.getHeaderSize();
                if (offset != mImageInput.getStreamPosition()) {
                    mImageInput.seek(offset);
                }
                ICOBitmapIndexed indexed = new ICOBitmapIndexed(entry);
                readColorMap(indexed);
                specifier = IndexedImageTypeSpecifier.createFromIndexColorModel(indexed.createColorModel());
                break;
            case 16:
                specifier = ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_USHORT_555_RGB);
                break;
            case 24:
                specifier = ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_3BYTE_BGR);
                break;
            case 32:
                specifier = ImageTypeSpecifier.createFromBufferedImageType(BufferedImage.TYPE_INT_ARGB);
                break;
            default:
                throw new IIOException("Unknown bit depth: " + header.getBitCount());


        }
        types.add(specifier);
        return types.iterator();
    }

    @Override
    public int getNumImages(boolean pAllowSearch) throws IOException {
        return getICOFile().getImageCount();
    }

    public int getWidth(int pImageIndex) throws IOException {
        return getICOEntry(pImageIndex).getWidth();
    }

    public BufferedImage read(int pImageIndex, ImageReadParam pParam) throws IOException {
        checkBounds(pImageIndex);

        processImageStarted(pImageIndex);

        BufferedImage image = getDestination(pParam, getImageTypes(pImageIndex), getWidth(pImageIndex), getHeight(pImageIndex));

        ICOEntry entry = getICOEntry(pImageIndex);
        initEntryHeader(entry);

        BufferedImage temp = readEntryBitmap(entry);

        // TODO: Handle AOI and subsampling inline
        if (pParam != null) {
            temp = fakeAOI(temp, pParam);
            temp = ImageUtil.toBuffered(fakeSubsampling(temp, pParam));
        }

        Graphics2D g = image.createGraphics();
        try {
            g.setComposite(AlphaComposite.Src);
            g.drawImage(temp, 0, 0, null);
        }
        finally {
            g.dispose();
        }

        processImageProgress(100);
        processImageComplete();

        return image;
//        return temp;
    }

    private ICOHeader initEntryHeader(ICOEntry pEntry) throws IOException {
        if (pEntry.getHeader() == null) {
            mImageInput.seek(pEntry.getOffset());
            ICOHeader header = new ICOHeader(mImageInput);
            pEntry.setHeader(header);
        }
        return pEntry.getHeader();
    }

    private BufferedImage readEntryBitmap(ICOEntry pEntry) throws IOException {
        if (pEntry.getBitmap() == null) {
            ICOHeader header = pEntry.getHeader();

            // We're usually at the exact spot here, reposition just in case
            int offset = pEntry.getOffset() + header.getHeaderSize();
            if (offset != mImageInput.getStreamPosition()) {
                mImageInput.seek(offset);
            }

            // TODO: Maybe we could support this, it's already in the BMP reader, check if spec allows!
            if (header.getCompression() != 0) {
                pEntry.setBitmap(new ICOBitmapUnsupported(pEntry, "Compression type not supported: " + header.getCompression()));
            }

            int bitCount = pEntry.getBitCount();

            ICOBitmap bitmap;
            switch (bitCount) {
                // Palette style
                case 1:
                case 4:
                case 8:
                    bitmap = new ICOBitmapIndexed(pEntry);
                    readBitmapIndexed((ICOBitmapIndexed) bitmap);
                    break;

                // RGB style
                case 16:
                    bitmap = new ICOBitmapRGB(pEntry);
                    readBitmap16(bitmap);
                    break;
                case 24:
                    bitmap = new ICOBitmapRGB(pEntry);
                    readBitmap24(bitmap);
                    break;
                case 32:
                    bitmap = new ICOBitmapRGB(pEntry);
                    readBitmap32(bitmap);
                    break;

                default:
                    bitmap = new ICOBitmapUnsupported(pEntry, "Unsupported bit count " + bitCount);
            }

            pEntry.setBitmap(bitmap);
        }

        return pEntry.getImage();
    }

    private void readBitmapIndexed(final ICOBitmapIndexed pBitmap) throws IOException {
        readColorMap(pBitmap);

        switch (pBitmap.mEntry.getBitCount()) {
            case 1:
                readBitmapIndexed1(pBitmap, false);
                break;
            case 4:
                readBitmapIndexed4(pBitmap);
                break;
            case 8:
                readBitmapIndexed8(pBitmap);
                break;
        }

        ICOBitmapMask mask = new ICOBitmapMask(pBitmap.mEntry);
        readBitmapIndexed1(mask.mMask, true);
        pBitmap.mEntry.setMask(mask);
    }

    private void readColorMap(ICOBitmapIndexed pBitmap) throws IOException {
        for (int i = 0; i < pBitmap.mEntry.getColorCount(); i++) {
            // aRGB (a is "Reserved")
            pBitmap.mColors[i] = (mImageInput.readInt() & 0xffffff) | 0xff000000;
        }
    }

    private void readBitmapIndexed1(ICOBitmapIndexed pBitmap, boolean pAsMask) throws IOException {
        int width = adjustToPadding(pBitmap.mWidth >> 3);
        byte[] row = new byte[width];

        for (int y = 0; y < pBitmap.mHeight; y++) {
            mImageInput.readFully(row, 0, width);
            int rowPos = 0;
            int xOrVal = 0x80;
            int pos = (pBitmap.mHeight - y - 1) * pBitmap.mWidth;

            for (int x = 0; x < pBitmap.mWidth; x++) {
                pBitmap.mBits[pos++] = ((row[rowPos] & xOrVal) / xOrVal) & 0xFF;
                if (xOrVal == 1) {
                    xOrVal = 0x80;
                    rowPos++;
                }
                else {
                    xOrVal >>= 1;
                }
            }

            // NOTE: If we are reading the mask, we don't abort or progress
            if (!pAsMask) {
                if (abortRequested()) {
                    processReadAborted();
                    break;
                }
                processImageProgress(100 * y / (float) pBitmap.mHeight);
            }
        }
    }

    private void readBitmapIndexed4(ICOBitmapIndexed pBitmap) throws IOException {
        int width = adjustToPadding(pBitmap.mWidth >> 1);
        byte[] row = new byte[width];

        for (int y = 0; y < pBitmap.mHeight; y++) {
            mImageInput.readFully(row, 0, width);
            int rowPos = 0;
            boolean high4 = true;
            int pos = (pBitmap.mHeight - y - 1) * pBitmap.mWidth;
            for (int x = 0; x < pBitmap.mWidth; x++) {
                int value;
                if (high4) {
                    value = (row[rowPos] & 0xF0) >> 4;
                }
                else {
                    value = row[rowPos] & 0x0F;
                    rowPos++;
                }
                pBitmap.mBits[pos++] = value & 0xFF;
                high4 = !high4;
            }

            if (abortRequested()) {
                processReadAborted();
                break;
            }
            processImageProgress(100 * y / (float) pBitmap.mHeight);
        }
    }

    private void readBitmapIndexed8(ICOBitmapIndexed pBitmap) throws IOException {
        int width = adjustToPadding(pBitmap.mWidth);

        byte[] row = new byte[width];

        for (int y = 0; y < pBitmap.mHeight; y++) {
            mImageInput.readFully(row, 0, width);
            int rowPos = 0;
            int pos = (pBitmap.mHeight - y - 1) * pBitmap.mWidth;

            for (int x = 0; x < pBitmap.mWidth; x++) {
                pBitmap.mBits[pos++] = row[rowPos++] & 0xFF;
            }

            if (abortRequested()) {
                processReadAborted();
                break;
            }
            processImageProgress(100 * y / (float) pBitmap.mHeight);
        }
    }

    /**
     * @param pWidth Bytes per scan line (i.e., 1BPP, width = 9 -> bytes = 1)
     * @return padded width
     */
    private static int adjustToPadding(final int pWidth) {
        if ((pWidth & 0x03) != 0) {
             return (pWidth & ~0x03) + 4;
        }
        return pWidth;
    }

    private void readBitmap16(final ICOBitmap pBitmap) throws IOException {
        // TODO: No idea if this actually works..
        // TODO: Use the last bit for aplha mask? That will cause TYPE_CUSTOM
        short[] pixels = new short[pBitmap.mWidth * pBitmap.mHeight];

        // Will create TYPE_USHORT_555;
        DirectColorModel cm = new DirectColorModel(16, 0x7C00, 0x03E0, 0x001F);
        DataBuffer buffer = new DataBufferShort(pixels, pixels.length);
        WritableRaster raster = Raster.createPackedRaster(buffer, pBitmap.mWidth, pBitmap.mHeight, pBitmap.mWidth, cm.getMasks(), null);
        pBitmap.mImage = new BufferedImage(cm, raster, cm.isAlphaPremultiplied(), null);       

        for (int y = 0; y < pBitmap.mHeight; y++) {
            int pos = (pBitmap.mHeight - y - 1) * pBitmap.mWidth;
            for (int x = 0; x < pBitmap.mWidth; x++) {
                // BGR -> RGB
                pixels[pos++] = (short) (mImageInput.readUnsignedShort() & 0xffff);
            }

            // Skip to 32 bit boundary
            if (pBitmap.mWidth % 2 != 0) {
                mImageInput.readUnsignedShort();
            }

            if (abortRequested()) {
                processReadAborted();
                break;
            }
            processImageProgress(100 * y / (float) pBitmap.mHeight);
        }
    }

    private void readBitmap24(final ICOBitmap pBitmap) throws IOException {
        byte[] pixels = new byte[pBitmap.mWidth * pBitmap.mHeight * 3];

        // Create TYPE_3BYTE_BGR
        DataBuffer buffer1 = new DataBufferByte(pixels, pixels.length);
        ColorSpace cs = ColorSpace.getInstance(ColorSpace.CS_sRGB);
        int[] nBits = {8, 8, 8};
        int[] bOffs = {2, 1, 0};
        ComponentColorModel cm = new ComponentColorModel(cs, nBits, false, false,
                                                         Transparency.OPAQUE,
                                                         DataBuffer.TYPE_BYTE);

        WritableRaster raster = Raster.createInterleavedRaster(buffer1, pBitmap.mWidth, pBitmap.mHeight, pBitmap.mWidth, 3, bOffs, null);
        pBitmap.mImage = new BufferedImage(cm, raster, cm.isAlphaPremultiplied(), null);
        
        for (int y = 0; y < pBitmap.mHeight; y++) {
            int pos = (pBitmap.mHeight - y - 1) * pBitmap.mWidth;
            mImageInput.readFully(pixels, pos, pBitmap.mWidth * 3);

            // TODO: Possibly read padding byte here!

            if (abortRequested()) {
                processReadAborted();
                break;
            }
            processImageProgress(100 * y / (float) pBitmap.mHeight);
        }
    }

    private void readBitmap32(final ICOBitmap pBitmap) throws IOException {
        int[] pixels = new int[pBitmap.mWidth * pBitmap.mHeight];

        // Will create TYPE_INT_ARGB
        DirectColorModel cm = (DirectColorModel) ColorModel.getRGBdefault();
        DataBuffer buffer = new DataBufferInt(pixels, pixels.length);
        WritableRaster raster = Raster.createPackedRaster(buffer, pBitmap.mWidth, pBitmap.mHeight, pBitmap.mWidth, cm.getMasks(), null);
        pBitmap.mImage = new BufferedImage(cm, raster, cm.isAlphaPremultiplied(), null);

        for (int y = 0; y < pBitmap.mHeight; y++) {
            int pos = (pBitmap.mHeight - y - 1) * pBitmap.mWidth;
            for (int x = 0; x < pBitmap.mWidth; x++) {
                // BGRA -> ARGB
                pixels[pos++] = mImageInput.readInt(); // This is ARGB when read LittleEndian
            }

            if (abortRequested()) {
                processReadAborted();
                break;
            }            
            processImageProgress(100 * y / (float) pBitmap.mHeight);
        }
    }

    /**
     * Get ICOFile object (cached).
     *
     * @return The ICOFile object
     * @throws java.io.IOException if an
     */
    private ICOFile getICOFile() throws IOException {
        assertInput();
        if (mICOFile == null) {
            mICOFile = new ICOFile(mImageInput);
        }

        return mICOFile;
    }

    private ICOEntry getICOEntry(int pImageIndex) throws IOException {
        ICOFile file = getICOFile();
        if (pImageIndex < 0 || pImageIndex >= file.getImageCount()) {
            throw new IndexOutOfBoundsException("Index: " + pImageIndex + ", ImageCount: " + file.getImageCount());
        }

        return file.getEntry(pImageIndex);
    }
    
    /// Test code below, ignore.. :-)
    public static void main(String[] pArgs) throws IOException {
        if (pArgs.length == 0) {
            System.err.println("Please specify the icon file name");
            System.exit(1);
        }

        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        }
        catch (Exception e) {
            // Ignore
        }

        final String fileName = pArgs[0];
        final File file = new File(fileName);
        if (!file.isFile()) {
            System.err.println(fileName + " not found, or is no file");
            System.exit(1);
        }

        final String title = file.getName();
        JFrame frame = createWindow(title);

        ImageReader reader = ImageIO.getImageReadersByFormatName("ico").next();
        if (reader == null) {
            System.err.println("No reader for format 'ico' found");
            System.exit(1);
        }
        reader.setInput(ImageIO.createImageInputStream(file));
        final int imageCount = reader.getNumImages(true);
        for (int i = 0; i < imageCount; i++) {
            try {
                addImage(frame.getContentPane(), reader, i);
            }
            catch (Exception e) {
                e.printStackTrace();
            }
        }

        frame.pack();
        frame.setVisible(true);
    }

    private static JFrame createWindow(final String pTitle) {
        JFrame frame = new JFrame(pTitle);
        frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
        frame.addWindowListener(new WindowAdapter() {
            public void windowClosed(WindowEvent e) {
                System.exit(0);
            }
        });
        frame.getContentPane().setLayout(
                new BoxLayout(frame.getContentPane(), BoxLayout.Y_AXIS)
        );
        return frame;
    }

    private static void addImage(Container pParent, ImageReader pReader, int pImageNo) throws IOException {
        final JButton button = new JButton();
        BufferedImage image = pReader.read(pImageNo);
        button.setIcon(new ImageIcon(image));
        button.setText("" + image.getWidth() + "x" +
                image.getHeight() + ": "
                + ((image.getColorModel() instanceof IndexColorModel) ?
                "" + ((IndexColorModel) image.getColorModel()).getMapSize() :
                "TrueColor"));
        pParent.add(button);
    }   
}
